#
# System Interface Library for games
# Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
# Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
# See the file COPYING.txt for details.
#
# build/linux/build.mk: Build control file for Linux.
#

#
# Linux-specific notes:
#
# - An additional rule, "test-coverage", is provided to help in checking
#   code coverage.  Running "make test-coverage" will build the executable,
#   run the built-in tests, and output code coverage data to "coverage.dat".
#   See the "Built-in tests" section of the README file for details.
#
# - The $(SIL_COVERAGE_TESTS) variable can be set to a comma-separated list
#   of test names to select which tests are to be run for test-coverage.
#
# -------------------------------------------------------------------------
#
# To use this file, include it in your Makefile after defining values for
# configuration variables and source lists, but before defining any rules.
# For example:
#
# PROGRAM_NAME = MyProgram
# EXECUTABLE_NAME = my-program
#
# MY_SOURCES = my-program.c
# MY_OBJECTS_C = $(MY_SOURCES:$(TOPDIR)/%.c=$(OBJDIR)/%$(OBJECT_EXT))
# OBJECT_GROUPS += MY
# AUTOGEN_HEADERS += my-date.h
#
# include SIL/build/linux/build.mk
#
# .PHONY: my-date.h  # Force regeneration on every build.
# my-date.h:
# 	date +'#define DATE "%Y-%m-%d %H:%M:%S"' >$@
#

###########################################################################
################ Defaults for common configuration options ################
###########################################################################

# Look for Valgrind installed in /usr, and use it if found.
_SIL_HAVE_VALGRIND := $(if $(wildcard /usr/include/valgrind/memcheck.h),1,0)
SIL_DEBUG_USE_VALGRIND ?= $(_SIL_HAVE_VALGRIND)

# Enable the WebM software decoder if FFmpeg is excluded.
SIL_MOVIE_INCLUDE_WEBM ?= $(if $(filter 1,$(USE_FFMPEG)),0,1)

###########################################################################
################## Linux-specific configuration options ###################
###########################################################################

#----------------------- Program behavior options ------------------------#

# $(SIL_SDL_CENTER_WINDOW):  If set to 1, any windows opened by the program
# will be centered on the screen.  If 0 (the default), the system's default
# behavior will be used.

SIL_SDL_CENTER_WINDOW ?= 0


# $(USE_FFMPEG):  Setting this to 1 enables use of the FFmpeg libraries
# (libavcodec, libavformat, libavresample, and libavutil) for decoding
# movie data.  If this is disabled, the libvpx-based WebM software decoder
# will be enabled by default.
#
# Currently, SIL requires FFmpeg 1.2 or later.  Note that an executable
# built against a particular version of FFmpeg generally requires the user
# to have a version from the same minor series (e.g., 1.2.x if built
# against 1.2.6) installed at runtime.

USE_FFMPEG ?= 0


# $(USE_SDL2):  Set this to 1 to build against SDL 2.0 only; set it to 0
# to build against SDL 1.2 only.  If empty (the default), the build will
# use SDL 1.2 if available and fall back to SDL 2.0 otherwise.

USE_SDL2 ?=

###########################################################################
######################### Build environment setup #########################
###########################################################################

_THISDIR := $(abspath $(dir $(lastword $(MAKEFILE_LIST))))

include $(_THISDIR)/../common/base.mk
include $(_THISDIR)/../common/config.mk

CC ?= gcc

OBJECT_EXT  = .o
LIBRARY_EXT = .a

SYS_CFLAGS    =
SYS_CXXFLAGS  =
SYS_OBJCFLAGS =
SYS_LDFLAGS   =

# $(-lSDL) is defined below.
BASE_LIBS = $(if $(filter 1,$(USE_FFMPEG)),-lavcodec -lavformat -lavresample -lavutil) \
            $(-lSDL) -lX11 -lasound -lm -lpthread -lrt

include $(_THISDIR)/../common/toolchain.mk

#-------------------------------------------------------------------------#

# Figure out which version of SDL to use.

_SIL_SDL1_HEADER_PATH := $(if $(filter 1,$(USE_SDL2)),,$(shell echo '\#include <SDL/SDL.h>' | $(CC) $(CFLAGS) $(CFLAG_PREPROCESS) - 2>/dev/null | grep '\# [1-9][0-9]* "[^"]*/SDL.h"' | head -n1 | sed -e 's|^[^"]*"\([^"]*\)/SDL\.h".*|\1|'))
_SIL_SDL2_HEADER_PATH := $(if $(filter 0,$(USE_SDL2)),,$(shell echo '\#include <SDL2/SDL.h>' | $(CC) $(CFLAGS) $(CFLAG_PREPROCESS) - 2>/dev/null | grep '\# [1-9][0-9]* "[^"]*/SDL.h"' | head -n1 | sed -e 's|^[^"]*"\([^"]*\)/SDL\.h".*|\1|'))
# Make sure we got an existing directory (and not a junk line).
_SIL_SDL1_HEADER_PATH := $(if $(shell ls -d '$(_SIL_SDL1_HEADER_PATH)' 2>/dev/null),$(_SIL_SDL1_HEADER_PATH))
_SIL_SDL2_HEADER_PATH := $(if $(shell ls -d '$(_SIL_SDL2_HEADER_PATH)' 2>/dev/null),$(_SIL_SDL2_HEADER_PATH))

ifneq ($(_SIL_SDL1_HEADER_PATH),)
    _SIL_SDL_HEADER_PATH := $(_SIL_SDL1_HEADER_PATH)
    -lSDL := -lSDL
else ifneq ($(_SIL_SDL2_HEADER_PATH),)
    _SIL_SDL_HEADER_PATH := $(_SIL_SDL2_HEADER_PATH)
    -lSDL := -lSDL2
else
    $(error Couldn't find SDL$(if $(filter 1,$(USE_SDL2)), 2.0,$(if $(filter 0,$(USE_SDL2)), 1.2)) headers)
endif

###########################################################################
################# Source file lists and associated flags ##################
###########################################################################

SIL_SYS_CONFIG_HEADER = linux.h

SIL_SYS_SOURCES = $(SIL_OPENGL_SOURCES) \
                  sysdep/linux/debug.c \
                  sysdep/linux/main.c \
                  sysdep/linux/meminfo.c \
                  sysdep/linux/misc.c \
                  sysdep/linux/movie.c \
                  sysdep/linux/sound.c \
                  sysdep/linux/sysfont.c \
                  sysdep/linux/userdata.c \
                  sysdep/posix/files.c \
                  sysdep/posix/fileutil.c \
                  sysdep/posix/ioqueue.c \
                  sysdep/posix/misc.c \
                  sysdep/posix/mutex.c \
                  sysdep/posix/semaphore.c \
                  sysdep/posix/thread.c \
                  sysdep/posix/time.c \
                  sysdep/posix/userdata.c \
                  sysdep/sdl/dyngl.c \
                  sysdep/sdl/graphics.c \
                  sysdep/sdl/input.c \
                  test/sysdep/linux/misc.c \
                  test/sysdep/linux/movie.c \
                  test/sysdep/linux/sound.c \
                  test/sysdep/posix/files.c \
                  test/sysdep/posix/fileutil.c \
                  test/sysdep/posix/internal.c \
                  test/sysdep/posix/ioqueue.c \
                  test/sysdep/posix/misc.c \
                  test/sysdep/posix/thread.c \
                  test/sysdep/posix/time.c \
                  test/sysdep/posix/userdata.c \
                  test/sysdep/sdl/graphics.c \
                  test/sysdep/sdl/input.c

# System-specific intenal SIL flags.  We set 32-byte alignment for texture
# buffers to improve copy performance.
SIL_SYS_FLAGS = $(CFLAG_DEFINE)SIL_OPENGL_TEXTURE_BUFFER_ALIGNMENT=32 \
                $(call _if-define,SIL_SDL_CENTER_WINDOW) \
                $(call _if-define,USE_FFMPEG,SIL_PLATFORM_LINUX_USE_FFMPEG) \
                $(CFLAG_INCLUDE_DIR)'$(_SIL_SDL_HEADER_PATH)'

# Declare that we follow POSIX.
SIL_SYS_FLAGS += $(CFLAG_DEFINE)_POSIX_C_SOURCE=201112L

include $(_THISDIR)/../common/sources.mk

###########################################################################
############################### Build rules ###############################
###########################################################################

SIL_DEFAULT_TARGET = all

SIL_HELP_all           = 'build the executable ($(EXECUTABLE_NAME))'
SIL_HELP_test-coverage = 'build and run tests, writing coverage data to coverage.out'
SIL_HELP_clean         = 'remove all intermediate files and dependency data'
SIL_HELP_spotless      = 'remove all generated files, including the executable'

include $(_THISDIR)/../common/rules.mk

#-------------------------------------------------------------------------#

.PHONY: all clean spotless

all: $(EXECUTABLE_NAME)

clean:
	$(ECHO) 'Removing object files'
	$(Q)rm -rf obj
	$(ECHO) 'Removing temporary files'
	$(Q)rm -rf .covtmp
	$(SIL_RECIPE_clean)

spotless: clean
	$(ECHO) 'Removing executable file'
	$(Q)rm -f '$(EXECUTABLE_NAME)'
	$(ECHO) 'Removing coverage data'
	$(Q)rm -rf coverage.out
	$(SIL_RECIPE_spotless)

#-------------------------------------------------------------------------#

.PHONY: test-coverage

test-coverage:
	$(Q)$(MAKE) all COVERAGE=1
	$(ECHO) 'Running tests'
	$(Q)find obj -name \*.gcda -delete
	$(Q)'./$(EXECUTABLE_NAME)' -test$(if $(SIL_COVERAGE_TESTS),=$(SIL_COVERAGE_TESTS))
	$(Q)$(MAKE) gen-coverage

#-------------------------------------------------------------------------#

# Sort objects in lexical order to make it easier to estimate build progress.
$(EXECUTABLE_NAME): $(sort $(ALL_OBJECTS))
	$(ECHO) 'Linking $@'
	$(Q)mkdir -p '$(@D)'
	$(Q)$(CXX) \
	    $(BASE_LDFLAGS) $(ALL_LDFLAGS) $(LDFLAGS) \
	    $(ALL_OBJECTS) \
	    $(ALL_LIBS) $(BASE_LIBS) $(LIBS) \
	    -o '$@'

###########################################################################
###########################################################################
