/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/test/utility/id-array.c: Tests for the ID array management routines.
 */

#ifdef INCLUDE_TESTS  // To the end of the file.

#include "src/base.h"
#include "src/memory.h"
#include "src/test/base.h"
#include "src/utility/id-array.h"

/*************************************************************************/
/*************************************************************************/

DEFINE_GENERIC_TEST_RUNNER(test_utility_id_array)

/*------------------------------------------------------------------------*/

TEST(test_basic)
{
    void * const object = (void *)0x1234;
    IDArray array = ID_ARRAY_INITIALIZER(10);
    int id;

    CHECK_TRUE(id = id_array_register(&array, object));
    CHECK_PTREQUAL(id_array_get(&array, id), object);
    CHECK_INTRANGE(id_array_size(&array), 10, 19);

    id_array_release(&array, id);
    return 1;
}

/*------------------------------------------------------------------------*/

TEST(test_memory_failure)
{
    void * const object = (void *)0x1234;
    IDArray array = ID_ARRAY_INITIALIZER(10);
    int id;

    CHECK_MEMORY_FAILURES(id = id_array_register(&array, object));
    CHECK_PTREQUAL(id_array_get(&array, id), object);

    id_array_release(&array, id);
    CHECK_FALSE(id_array_get(&array, id));

    return 1;
}

/*------------------------------------------------------------------------*/

TEST(test_expand_array)
{
    char * const object = (void *)0x1234;
    IDArray array = ID_ARRAY_INITIALIZER(10);
    int id[11];

    for (int i = 0; i < 11; i++) {
        CHECK_TRUE(id[i] = id_array_register(&array, object + i));
    }
    for (int i = 0; i < 11; i++) {
        CHECK_PTREQUAL(id_array_get(&array, id[i]), object + i);
    }
    CHECK_INTRANGE(id_array_size(&array), 20, 29);

    for (int i = 0; i < 11; i++) {
        id_array_release(&array, id[i]);
    }
    return 1;
}

/*------------------------------------------------------------------------*/

TEST(test_id_reuse)
{
    void * const object1 = (void *)0x1234;
    void * const object2 = (void *)0x5678;
    void * const object3 = (void *)0xABCD;
    void * const object4 = (void *)0x4321;
    void * const object5 = (void *)0x8765;
    IDArray array = ID_ARRAY_INITIALIZER(10);
    int id1, id2, id3;

    CHECK_TRUE(id1 = id_array_register(&array, object1));
    CHECK_TRUE(id2 = id_array_register(&array, object2));
    CHECK_TRUE(id3 = id_array_register(&array, object3));
    CHECK_TRUE(id1 < id2);
    CHECK_TRUE(id2 < id3);

    id_array_release(&array, id1);
    id_array_release(&array, id2);
    CHECK_INTEQUAL(id_array_register(&array, object4), id1);
    CHECK_INTEQUAL(id_array_register(&array, object5), id2);
    CHECK_PTREQUAL(id_array_get(&array, id1), object4);
    CHECK_PTREQUAL(id_array_get(&array, id2), object5);
    CHECK_PTREQUAL(id_array_get(&array, id3), object3);

    id_array_release(&array, id1);
    id_array_release(&array, id2);
    id_array_release(&array, id3);
    return 1;
}

/*------------------------------------------------------------------------*/

TEST(test_get_id_out_of_range)
{
    void * const object = (void *)0x1234;
    IDArray array = ID_ARRAY_INITIALIZER(10);
    int id;

    CHECK_TRUE(id = id_array_register(&array, object));
    CHECK_PTREQUAL(id_array_get(&array, id), object);
    CHECK_FALSE(id_array_get(&array, 0));
    CHECK_FALSE(id_array_get(&array, 11));

    id_array_release(&array, id);
    return 1;
}

/*------------------------------------------------------------------------*/

TEST(test_release_id_out_of_range)
{
    void * const object = (void *)0x1234;
    IDArray array = ID_ARRAY_INITIALIZER(10);
    int id;

    CHECK_TRUE(id = id_array_register(&array, object));
    id_array_release(&array, 0);
    id_array_release(&array, 11);
    CHECK_PTREQUAL(id_array_get(&array, id), object);

    id_array_release(&array, id);
    return 1;
}

/*------------------------------------------------------------------------*/

TEST(test_initially_empty)
{
    IDArray array = ID_ARRAY_INITIALIZER(10);
    CHECK_FALSE(id_array_get(&array, 1));
    return 1;
}

/*************************************************************************/
/*************************************************************************/

#endif  // INCLUDE_TESTS

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
