/*
 * System Interface Library for games
 * Copyright (c) 2007-2014 Andrew Church <achurch@achurch.org>
 * Released under the GNU GPL version 3 or later; NO WARRANTY is provided.
 * See the file COPYING.txt for details.
 *
 * src/utility/yuv2rgb.c: YUV/RGB colorspace conversion functionality.
 */

/*
 * Performance data for x86_64 using an Intel Core i7 4770S: (times are in
 * millions of RDTSC ticks for a 640x480 video frame)
 *             | C version | SSE2 version
 * ------------+-----------+-------------
 *  Unsmoothed |    18.0   |     0.8
 *    Smoothed |    28.3   |     3.5
 *
 * Performance data for MIPS on a PSP running at 222 MHz:
 *    C version:        78285 used / 640x480 frame
 *    Assembly version: 58825 used / 640x480 frame
 */

#include "src/base.h"
#include "src/utility/yuv2rgb.h"

#ifdef SIL_ARCH_X86
# include "src/utility/asm-x86.h"
#endif

/*************************************************************************/
/****************************** Local data *******************************/
/*************************************************************************/

/**
 * YUVConversionFunc:  Type of a function to convert a row of pixels from
 * YUV to RGBA format.
 *
 * [Parameters]
 *     srcY: Pointer to the start of the row in the Y plane of the source data.
 *     srcU: Pointer to the start of the row in the U plane of the source data.
 *     srcV: Pointer to the start of the row in the V plane of the source data.
 *     dest: Pointer to the start of the row in the destination buffer.
 *     width: Number of pixels to convert.
 */
typedef void YUVConversionFunc(
    const uint8_t *srcY, const uint8_t *srcU, const uint8_t *srcV,
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
    const uint8_t *srcU2, const uint8_t *srcV2,
#endif
    uint8_t *dest, int width);


/* Conversion functions. */

static YUVConversionFunc convert_c;

#if defined(SIL_ARCH_MIPS) && !defined(SIL_UTILITY_YUV2RGB_SMOOTH_UV) && (defined(__GNUC__) || defined(__clang__))
static YUVConversionFunc convert_mips;
#endif

#if defined(SIL_ARCH_X86) && (defined(__GNUC__) || defined(__clang__))
static YUVConversionFunc convert_sse2;
#endif

/*************************************************************************/
/************************** Interface routines ***************************/
/*************************************************************************/

void yuv2rgb(const uint8_t **src, const int *src_stride,
             uint8_t *dest, const int dest_stride,
             const int width, const int height)
{
    /* Check that parameters are sane. */
    PRECOND(src != NULL, return);
    PRECOND(src[0] != NULL, return);
    PRECOND(src[1] != NULL, return);
    PRECOND(src[2] != NULL, return);
    PRECOND(src_stride != NULL, return);
    PRECOND(src_stride[0] > 0);
    PRECOND(src_stride[1] > 0);
    PRECOND(src_stride[2] > 0);
    PRECOND(dest != NULL, return);
    PRECOND(dest_stride > 0);
    PRECOND(width > 0);
    PRECOND(height > 0);

    /* Pick a conversion routine to use based on the availability of an
     * optimized routine for the current architecture and the alignment of
     * the parameter values. */
    YUVConversionFunc *convert = convert_c;
#if defined(SIL_ARCH_MIPS) && !defined(SIL_UTILITY_YUV2RGB_SMOOTH_UV) && (defined(__GNUC__) || defined(__clang__))
    if (((intptr_t)dest & 3) == 0) {
        convert = convert_mips;
    }
#endif
#if defined(SIL_ARCH_X86) && (defined(__GNUC__) || defined(__clang__))
    if (((intptr_t)src[0] & 15) == 0
     && src_stride[0] % 16 == 0
     && ((intptr_t)dest & 15) == 0
     && dest_stride % 4 == 0
     && width % 16 == 0) {
        convert = convert_sse2;
    }
#endif

    /* Run the conversion function on each line of the image. */
    int y;
    for (y = 0; y < height; y++, dest += dest_stride*4) {
        const uint8_t * const srcY = src[0] + y * src_stride[0];
        const uint8_t * const srcU = src[1] + (y/2) * src_stride[1];
        const uint8_t * const srcV = src[2] + (y/2) * src_stride[2];
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        const uint8_t * const srcU2 =
            (y%2==0 ? (y==0 ? srcU : srcU - src_stride[1])
                    : (y==height-1 ? srcU : srcU + src_stride[1]));
        const uint8_t * const srcV2 =
            (y%2==0 ? (y==0 ? srcV : srcV - src_stride[2])
                    : (y==height-1 ? srcV : srcV + src_stride[2]));
#endif

        (*convert)(srcY, srcU, srcV,
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
                   srcU2, srcV2,
#endif
                   dest, width);
    }
}

/*************************************************************************/
/******************* Single-line conversion functions ********************/
/*************************************************************************/

static void convert_c(
    const uint8_t *srcY, const uint8_t *srcU, const uint8_t *srcV,
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
    const uint8_t *srcU2, const uint8_t *srcV2,
#endif
    uint8_t *dest, int width)
{
    int x;
    for (x = 0; x < width/2; x++) {
        int Y, U, V, R, G, B;
        Y = (srcY[x*2] - 0x10) * 9539;
        U = srcU[x];
        V = srcV[x];
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        int thisU = U;
        int thisV = V;
        int lastU = x==0 ? thisU : srcU[x-1];
        int thisU2 = srcU2[x];
        int lastU2 = x==0 ? thisU2 : srcU2[x-1];
        U = (thisU*9 + lastU*3 + thisU2*3 + lastU2*1) / 16;
        int lastV = x==0 ? thisV : srcV[x-1];
        int thisV2 = srcV2[x];
        int lastV2 = x==0 ? thisV2 : srcV2[x-1];
        V = (thisV*9 + lastV*3 + thisV2*3 + lastV2*1) / 16;
#endif
        U -= 0x80;
        V -= 0x80;
        R = (Y           + 13075*V + (1<<12)) >> 13;
        G = (Y -  3209*U -  6660*V + (1<<12)) >> 13;
        B = (Y + 16525*U           + (1<<12)) >> 13;
        dest[x*8+0] = bound(R, 0, 255);
        dest[x*8+1] = bound(G, 0, 255);
        dest[x*8+2] = bound(B, 0, 255);
        dest[x*8+3] = 0xFF;
        Y = (srcY[x*2+1] - 0x10) * 9539;
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        int nextU = x==width/2-1 ? thisU : srcU[x+1];
        int nextU2 = x==width/2-1 ? thisU2 : srcU2[x+1];
        U = (thisU*9 + nextU*3 + thisU2*3 + nextU2*1) / 16;
        int nextV = x==width/2-1 ? thisV : srcV[x+1];
        int nextV2 = x==width/2-1 ? thisV2 : srcV2[x+1];
        V = (thisV*9 + nextV*3 + thisV2*3 + nextV2*1) / 16;
        U -= 0x80;
        V -= 0x80;
#endif
        R = (Y           + 13075*V + (1<<12)) >> 13;
        G = (Y -  3209*U -  6660*V + (1<<12)) >> 13;
        B = (Y + 16525*U           + (1<<12)) >> 13;
        dest[x*8+4] = bound(R, 0, 255);
        dest[x*8+5] = bound(G, 0, 255);
        dest[x*8+6] = bound(B, 0, 255);
        dest[x*8+7] = 0xFF;
    }
}

/*-----------------------------------------------------------------------*/

#if defined(SIL_ARCH_MIPS) && !defined(SIL_UTILITY_YUV2RGB_SMOOTH_UV) && (defined(__GNUC__) || defined(__clang__))

static void convert_mips(
    const uint8_t *srcY, const uint8_t *srcU, const uint8_t *srcV,
    uint8_t *dest, int width)
{
    /* Register usage:
     *    $at: Temporary
     *    $v0: RGBA output pixel
     *    $t0: Y pixel
     *    $t1: U pixel (raw, then scaled for blue)
     *    $t2: V pixel scaled for red
     *    $t3: U+V pixels scaled for green
     *    $t4: V pixel (raw), then temporary
     *    $t5: Y constant
     *    $t6: V constant for red
     *    $t7: U constant for green
     *    $t8: V constant for green
     *    $t9: U constant for blue
     *    $s0: Second Y pixel
     *    $s1: Temporary (non-PSP only)
     *    $s2: Temporary (non-PSP only)
     *    $s3: 255 (for bounding pixel values and setting output alpha value)
     */

    /* MINMAX:  Bound the value in $at to the range [0,255].  Uses $s1 and
     * $s2 as temporaries on platforms without the MIN and MAX instructions. */
    #ifdef SIL_PLATFORM_PSP
        #define MINMAX \
            "min $at, $at, $s3\n"   \
            "max $at, $at, $zero\n"
    #else
       #define MINMAX \
            "slt $s1, $at, $s3\n"   \
            "slt $s2, $at, $zero\n" \
            "movz $at, $s3, $s1\n"  \
            "movn $at, $zero, $s2\n"
    #endif

    #if defined(SIL_ARCH_MIPS32R2) || defined(SIL_PLATFORM_PSP)
        #define INS(n) \
            "ins $v0, $at, "#n", 8\n"
        #define INS_0 \
            "ins $v0, $at, 0, 8\n"
        #define CLEAR  /*nothing*/
    #else
        #define INS(n) \
            "sll $at, $at, "#n"\n" \
            "or $v0, $v0, $at\n"
        #define INS_0 \
            "or $v0, $v0, $at\n"
        #define CLEAR \
            "lui $v0, 0xFF00\n"
    #endif

    __asm__ volatile(
        ".set push; .set noreorder; .set noat   \n"
        "li $t5, 9539                           \n"
        "li $t6, 13075                          \n"
        "li $t7, -3209                          \n"
        "li $t8, -6660                          \n"
        "li $t9, 16525                          \n"
        "li $s3, 255                            \n"
        "lui $v0, 0xFF00                        \n"
        "0:                                     \n"
        "# Load, unbias, and scale              \n"
        "lbu $t4, 0(%[srcV])                    \n"
        "lbu $t0, 0(%[srcY])                    \n"
        "lbu $t1, 0(%[srcU])                    \n"
        "addiu $t4, $t4, -128                   \n"
        "mult $t4, $t6                          \n"
        "addiu %[srcU], %[srcU], 1              \n"
        "addiu %[srcV], %[srcV], 1              \n"
        "mflo $t2                               \n"
        "addiu $t0, $t0, -16                    \n"
        "lbu $s0, 1(%[srcY])                    \n"
        "mult $t0, $t5                          \n"
        "addiu %[width], %[width], -2           \n"
        "addiu %[srcY], %[srcY], 2              \n"
        "mflo $t0                               \n"
        "addiu $t1, $t1, -128                   \n"
        "        addu $at, $t0, $t2             \n"
        "mult $t1, $t7                          \n"
        "mflo $t3                               \n"
        "        addiu $at, $at, 4096           \n"
        "        sra $at, $at, 13               \n"
        "mult $t1, $t9                          \n"
        "        "MINMAX"                       \n"
        "        "CLEAR"                        \n"
        "        "INS_0"                        \n"
        "        or $v0, $v0, $at               \n"
        "mflo $t1                               \n"
        "        addu $at, $t0, $t1             \n"
        "        addiu $at, $at, 4096           \n"
        "mult $t4, $t8                          \n"
        "        sra $at, $at, 13               \n"
        "        "MINMAX"                       \n"
        "        "INS(16)"                      \n"
        "mflo $t4                               \n"
        "addu $t3, $t3, $t4                     \n"
        "addiu $s0, $s0, -16                    \n"
        "mult $s0, $t5                          \n"
        "        addu $at, $t0, $t3             \n"
        "        addiu $at, $at, 4096           \n"
        "        sra $at, $at, 13               \n"
        "        "MINMAX"                       \n"
        "        "INS(8)"                       \n"
        "        sw $v0, 0(%[dest])             \n"
        "                # Second output pixel  \n"
        "mflo $s0                               \n"
        "                addu $at, $s0, $t2     \n"
        "                addiu $at, $at, 4096   \n"
        "                sra $at, $at, 13       \n"
        "                "MINMAX"               \n"
        "                "CLEAR"                \n"
        "                "INS_0"                \n"
        "                addu $at, $s0, $t1     \n"
        "                addiu $at, $at, 4096   \n"
        "                sra $at, $at, 13       \n"
        "                "MINMAX"               \n"
        "                "INS(16)"              \n"
        "                addu $at, $s0, $t3     \n"
        "                addiu $at, $at, 4096   \n"
        "                sra $at, $at, 13       \n"
        "                "MINMAX"               \n"
        "                "INS(8)"               \n"
        "                sw $v0, 4(%[dest])     \n"
        "# Loop over remaining pixels           \n"
        "bnez %[width], 0b                      \n"
        "addiu %[dest], %[dest], 8              \n"
        ".set pop                               \n"
        : /* no outputs */
        : [srcY] "r" (srcY), [srcU] "r" (srcU), [srcV] "r" (srcV),
          [dest] "r" (dest), [width] "r" (width)
        : "at", "v0", "t0", "t1", "t2", "t3", "t4", "t5", "t6", "t7", "t8",
          "t9", "s0",
#ifndef SIL_PLATFORM_PSP
          "s1", "s2",
#endif
          "s3"
    );
}

#endif  // SIL_ARCH_MIPS && !SIL_UTILITY_YUV2RGB_SMOOTH_UV

/*-----------------------------------------------------------------------*/

#if defined(SIL_ARCH_X86) && (defined(__GNUC__) || defined(__clang__))

#if defined(__GNUC__) && !defined(__clang__) && !(__GNUC__ == 4 && __GNUC_MINOR__ >= 6 && __GNUC_MINOR__ <= 8)
/* Make sure we have enough registers for the inline asm block.
 * http://gcc.gnu.org/bugzilla/show_bug.cgi?id=60062 causes this to break
 * when compiling with -O0 on GCC 4.6 through 4.8, so for those compilers
 * we give up and save/restore another register instead. */
__attribute__((optimize(2,omit-frame-pointer)))
#endif
static void convert_sse2(
    const uint8_t *srcY, const uint8_t *srcU, const uint8_t *srcV,
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
    const uint8_t *srcU2, const uint8_t *srcV2,
#endif
    uint8_t *dest, int width)
{
    /* Constant data used in converting YUV to RGB. */
    static const struct { uint16_t n[80]; } ALIGNED(16) yuv_data = {{
        0x00FF,0x00FF,0x00FF,0x00FF,0x00FF,0x00FF,0x00FF,0x00FF, //for odd/even
        0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010,0x0010, //for Y -16
        0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080,0x0080, //for U/V -128
        0x2543,0x2543,0x2543,0x2543,0x2543,0x2543,0x2543,0x2543, //Y constant
        0x3313,0x3313,0x3313,0x3313,0x3313,0x3313,0x3313,0x3313, //rV constant
        0xF377,0xF377,0xF377,0xF377,0xF377,0xF377,0xF377,0xF377, //gU constant
        0xE5FC,0xE5FC,0xE5FC,0xE5FC,0xE5FC,0xE5FC,0xE5FC,0xE5FC, //gV constant
        0x408D,0x408D,0x408D,0x408D,0x408D,0x408D,0x408D,0x408D, //bU constant
        0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008,0x0008, //for rounding
        0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF,0xFFFF, //alpha values
    }};

    /* GCC is unable to align the stack properly, so we need to do it
     * ourselves. */
    struct srcYUV_struct {
        const uint8_t *Y, *U, *V;
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        const uint8_t *U2, *V2;
        uint8_t lastU, lastV, lastU2, lastV2;
        uint8_t pad[48-(PTRSIZE*5+4)];
        uint8_t tmp[16], Uright[16], Vright[16];
#endif
    };
    char srcYUV_buffer[96+15];
    ASSERT(sizeof(srcYUV_buffer)-15 >= sizeof(struct srcYUV_struct));
    struct srcYUV_struct *srcYUV =
        (struct srcYUV_struct *)align_up((uintptr_t)srcYUV_buffer, 16);
    srcYUV->Y = srcY;
    srcYUV->U = srcU;
    srcYUV->V = srcV;
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
    srcYUV->U2 = srcU2;
    srcYUV->V2 = srcV2;
    srcYUV->lastU = srcU[width/2-1];
    srcYUV->lastV = srcV[width/2-1];
    srcYUV->lastU2 = srcU2[width/2-1];
    srcYUV->lastV2 = srcV2[width/2-1];
#endif

    /* Dummy variable used to let GCC know that the ECX/RCX register is
     * changed by the assembly code (we pass a parameter in that register,
     * so we can't list it in the clobbers). */
    int dummy;

    __asm__ volatile(
#ifdef SIL_ARCH_X86_32
        "push %%ebx                      # Might be in use a a PIC register.\n"
# if defined(__GNUC__) && !defined(__clang__) && __GNUC__ == 4 && __GNUC_MINOR__ >= 6 && __GNUC_MINOR__ <= 8
        "push %%edx                      # GCC bug workaround.              \n"
# endif
#endif
        "lea %[yuv_data], "RBX"                                             \n"
        "0:                                                                 \n"
        "# Load data, bias and expand to 16 bits                            \n"
        "pxor %%xmm4, %%xmm4             # XMM4: 00 00 00 00 00 00 00 00    \n"
        "mov 0*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "movdqa ("RAX","RCX",2), %%xmm6  # XMM6: YF...................Y0    \n"
        "mov 1*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "movq ("RAX","RCX"), %%xmm2      # XMM2:             U7.......U0    \n"
        "mov 2*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "movq ("RAX","RCX"), %%xmm3      # XMM3:             V7.......V0    \n"
        "movdqa %%xmm6, %%xmm7           # XMM7: YF...................Y0    \n"
        "pand ("RBX"), %%xmm6            # XMM6: YE YC YA Y8 Y6 Y4 Y2 Y0    \n"
        "psrlw $8, %%xmm7                # XMM7: YF YD YB Y9 Y7 Y5 Y3 Y1    \n"
        "psubw 16("RBX"), %%xmm6         # XMM6: subtract 16                \n"
        "psllw $7, %%xmm6                # XMM6: convert to fixed point 8.7 \n"
        "psubw 16("RBX"), %%xmm7         # XMM7: subtract 16                \n"
        "psllw $7, %%xmm7                # XMM7: convert to fixed point 8.7 \n"
        "punpcklbw %%xmm4, %%xmm2        # XMM2: U7 U6 U5 U4 U3 U2 U1 U0    \n"
        "punpcklbw %%xmm4, %%xmm3        # XMM3: V7 V6 V5 V4 V3 V2 V1 V0    \n"
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        "# Smooth U and V values                                            \n"
        "mov $0x00030003, %%eax                                             \n"
        "movd %%eax, %%xmm4              # XMM4: -- -- -- -- -- --  3  3    \n"
        "pshufd $0, %%xmm4, %%xmm4       # XMM4:  3  3  3  3  3  3  3  3    \n"
        "movdqa 32("RBX"), %%xmm5        # XMM5: 1/2 ... 1/2 as fixed 8.0   \n"
        "psllw $4, %%xmm5                # XMM5: 1/2 ... 1/2 as fixed 8.4   \n"
        "lea -1("RCX"), "RDX"            # EDX: index of pixel to the left  \n"
        "test %%ecx, %%ecx               #   (careful to stay within buffer)\n"
        "cmovz %%ecx, %%edx                                                 \n"
        "# ((U[0:7]*3+U[1:8]) * 3 + (U2[0:7]*3+U2[1:8])) / 16 -> Uright     \n"
        "movd %%xmm2, %%eax              # AL: U0                           \n"
        "xchg 5*"PTRSIZE_STR"+0("RSI"), %%al  # AL <-> lastU                \n"
        "movd %%eax, %%xmm0              # XMM0: -- -- -- -- -- -- -- U8    \n"
        "movdqa %%xmm2, %%xmm1           # XMM1: U7 U6 U5 U4 U3 U2 U1 U0    \n"
        "pslldq $14, %%xmm0              # XMM0: U8 -- -- -- -- -- -- --    \n"
        "psrldq $2, %%xmm1               # XMM1: -- U7 U6 U5 U4 U3 U2 U1    \n"
        "por %%xmm0, %%xmm1              # XMM1: U8 U7 U6 U5 U4 U3 U2 U1    \n"
        "movdqa %%xmm2, %%xmm0           # XMM0: U7 U6 U5 U4 U3 U2 U1 U0    \n"
        "pmullw %%xmm4, %%xmm0           # XMM0: multiply by 3              \n"
        "paddw %%xmm1, %%xmm0            # XMM0: U[0:7]*3 + U[1:8]          \n"
        "pmullw %%xmm4, %%xmm0           # XMM0: multiply by 3              \n"
        "movdqa %%xmm0, 64("RSI")        # XMM0: flush to Uright            \n"
        "mov 3*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "pxor %%xmm0, %%xmm0                                                \n"
        "movq ("RAX","RCX"), %%xmm1                                         \n"
        "punpcklbw %%xmm0, %%xmm1        # XMM1: U2[0:7]                    \n"
        "movdqa %%xmm1, 48("RSI")        # XMM1: save in tmp                \n"
        "movd %%xmm1, %%eax              # AL: U2[0]                        \n"
        "xchg 5*"PTRSIZE_STR"+2("RSI"), %%al  # AL <-> lastU2               \n"
        "movd %%eax, %%xmm0              # XMM0: -- -- -- -- -- -- -- U2[8] \n"
        "pslldq $14, %%xmm0              # XMM0: U2[8] -- -- -- -- -- -- -- \n"
        "psrldq $2, %%xmm1               # XMM1: -- U2[7:1]                 \n"
        "por %%xmm0, %%xmm1              # XMM1: U2[8:1]                    \n"
        "movdqa 48("RSI"), %%xmm0        # XMM0: U2[7:0]                    \n"
        "pmullw %%xmm4, %%xmm0           # XMM0: multiply by 3              \n"
        "paddw %%xmm1, %%xmm0            # XMM0: U2[0:7]*3 + U2[1:8]        \n"
        "paddw 64("RSI"), %%xmm0         # XMM0: add (U[0:7]*3 + U[1:8]) * 3\n"
        "psubw %%xmm5, %%xmm0            # XMM0: subtract 0.5 @ fixed 8.4   \n"
        "psllw $3, %%xmm0                # XMM0: convert to fixed point 8.7 \n"
        "movdqa %%xmm0, 64("RSI")        # XMM0: store in Uright            \n"
        "# ((U[0:7]*3+U[-1:6]) * 3 + (U2[0:7]*3+U2[-1:6])) / 16 -> Uleft    \n"
        "mov 1*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "movzbl ("RAX","RDX"), %%eax                                        \n"
        "movd %%eax, %%xmm0              # XMM0: -- -- -- -- -- -- -- U-    \n"
        "movdqa %%xmm2, %%xmm1           # XMM1: U7 U6 U5 U4 U3 U2 U1 U0    \n"
        "pslldq $2, %%xmm1               # XMM1: U6 U5 U4 U3 U2 U1 U0 --    \n"
        "por %%xmm0, %%xmm1              # XMM1: U6 U5 U4 U3 U2 U1 U0 U-    \n"
        "pmullw %%xmm4, %%xmm2           # XMM2: multiply by 3              \n"
        "paddw %%xmm1, %%xmm2            # XMM2: U[0:7]*3 + U[-1:6]         \n"
        "pmullw %%xmm4, %%xmm2           # XMM2: multiply by 3              \n"
        "mov 3*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "movzbl ("RAX","RDX"), %%eax                                        \n"
        "movd %%eax, %%xmm0              # XMM0: -- -- -- -- -- -- -- U2[-1]\n"
        "movdqa 48("RSI"), %%xmm1        # XMM1: U2[7:0]                    \n"
        "pslldq $2, %%xmm1               # XMM1: U2[6:0] --                 \n"
        "por %%xmm0, %%xmm1              # XMM1: U2[6:-1]                   \n"
        "movdqa 48("RSI"), %%xmm0        # XMM0: U2[7:0]                    \n"
        "pmullw %%xmm4, %%xmm0           # XMM0: multiply by 3              \n"
        "paddw %%xmm1, %%xmm2            # XMM2: add U2[-1:6]               \n"
        "paddw %%xmm0, %%xmm2            # XMM2: add U2[0:7]*3              \n"
        "psubw %%xmm5, %%xmm2            # XMM2: subtract 0.5 @ fixed 8.4   \n"
        "psllw $3, %%xmm2                # XMM2: convert to fixed point 8.7 \n"
        "# ((V[0:7]*3+V[1:8]) * 3 + (V2[0:7]*3+V2[1:8])) / 16 -> Vright\n"
        "movd %%xmm3, %%eax              # AL: V0                           \n"
        "xchg 5*"PTRSIZE_STR"+1("RSI"), %%al  # AL <-> lastV                \n"
        "movd %%eax, %%xmm0              # XMM0: -- -- -- -- -- -- -- V8    \n"
        "movdqa %%xmm3, %%xmm1           # XMM1: V7 V6 V5 V4 V3 V2 V1 V0    \n"
        "pslldq $14, %%xmm0              # XMM0: V8 -- -- -- -- -- -- --    \n"
        "psrldq $2, %%xmm1               # XMM1: -- V7 V6 V5 V4 V3 V2 V1    \n"
        "por %%xmm0, %%xmm1              # XMM1: V8 V7 V6 V5 V4 V3 V2 V1    \n"
        "movdqa %%xmm3, %%xmm0           # XMM0: V7 V6 V5 V4 V3 V2 V1 V0    \n"
        "pmullw %%xmm4, %%xmm0           # XMM0: multiply by 3              \n"
        "paddw %%xmm1, %%xmm0            # XMM0: V[0:7]*3 + V[1:8]          \n"
        "pmullw %%xmm4, %%xmm0           # XMM0: multiply by 3              \n"
        "movdqa %%xmm0, 80("RSI")        # XMM0: flush to Vright            \n"
        "mov 4*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "pxor %%xmm0, %%xmm0                                                \n"
        "movq ("RAX","RCX"), %%xmm1                                         \n"
        "punpcklbw %%xmm0, %%xmm1        # XMM1: V2[0:7]                    \n"
        "movdqa %%xmm1, 48("RSI")        # XMM1: save in tmp                \n"
        "movd %%xmm1, %%eax              # AL: V2[0]                        \n"
        "xchg 5*"PTRSIZE_STR"+3("RSI"), %%al  # AL <-> lastV2               \n"
        "movd %%eax, %%xmm0              # XMM0: -- -- -- -- -- -- -- V2[8] \n"
        "pslldq $14, %%xmm0              # XMM0: V2[8] -- -- -- -- -- -- -- \n"
        "psrldq $2, %%xmm1               # XMM1: -- V2[7:1]                 \n"
        "por %%xmm0, %%xmm1              # XMM1: V2[8:1]                    \n"
        "movdqa 48("RSI"), %%xmm0        # XMM0: V2[7:0]                    \n"
        "pmullw %%xmm4, %%xmm0           # XMM0: multiply by 3              \n"
        "paddw %%xmm1, %%xmm0            # XMM0: V2[0:7]*3 + V2[1:8]        \n"
        "paddw 80("RSI"), %%xmm0         # XMM0: add (V[0:7]*3 + V[1:8]) * 3\n"
        "psubw %%xmm5, %%xmm0            # XMM0: subtract 0.5 @ fixed 8.4   \n"
        "psllw $3, %%xmm0                # XMM0: convert to fixed point 8.7 \n"
        "movdqa %%xmm0, 80("RSI")        # XMM0: store in Vright            \n"
        "# ((V[0:7]*3+V[-1:6]) * 3 + (V2[0:7]*3+V2[-1:6])) / 16 -> Vleft    \n"
        "mov 2*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "movzbl ("RAX","RDX"), %%eax                                        \n"
        "movd %%eax, %%xmm0              # XMM0: -- -- -- -- -- -- -- V-    \n"
        "movdqa %%xmm3, %%xmm1           # XMM1: V7 V6 V5 V4 V3 V2 V1 V0    \n"
        "pslldq $2, %%xmm1               # XMM1: V6 V5 V4 V3 V2 V1 V0 --    \n"
        "por %%xmm0, %%xmm1              # XMM1: V6 V5 V4 V3 V2 V1 V0 V-    \n"
        "pmullw %%xmm4, %%xmm3           # XMM3: multiply by 3              \n"
        "paddw %%xmm1, %%xmm3            # XMM3: V[0:7]*3 + V[-1:6]         \n"
        "pmullw %%xmm4, %%xmm3           # XMM3: multiply by 3              \n"
        "mov 4*"PTRSIZE_STR"("RSI"), "RAX"                                  \n"
        "movzbl ("RAX","RDX"), %%eax                                        \n"
        "movd %%eax, %%xmm0              # XMM0: -- -- -- -- -- -- -- V2[-1]\n"
        "movdqa 48("RSI"), %%xmm1        # XMM1: V2[7:0]                    \n"
        "pslldq $2, %%xmm1               # XMM1: V2[6:0] --                 \n"
        "por %%xmm0, %%xmm1              # XMM1: V2[6:-1]                   \n"
        "movdqa 48("RSI"), %%xmm0        # XMM0: V2[7:0]                    \n"
        "pmullw %%xmm4, %%xmm0           # XMM0: multiply by 3              \n"
        "paddw %%xmm1, %%xmm3            # XMM3: add V2[-1:6]               \n"
        "paddw %%xmm0, %%xmm3            # XMM3: add V2[0:7]*3              \n"
        "psubw %%xmm5, %%xmm3            # XMM3: subtract 0.5 @ fixed 8.4   \n"
        "psllw $3, %%xmm3                # XMM3: convert to fixed point 8.7 \n"
#else  // !SIL_UTILITY_YUV2RGB_SMOOTH_UV
        "psubw 32("RBX"), %%xmm2         # XMM2: subtract 128               \n"
        "psllw $7, %%xmm2                # XMM2: convert to fixed point 8.7 \n"
        "psubw 32("RBX"), %%xmm3         # XMM3: subtract 128               \n"
        "psllw $7, %%xmm3                # XMM3: convert to fixed point 8.7 \n"
#endif
        "# Multiply by constants                                            \n"
        "pmulhw 48("RBX"), %%xmm6        # XMM6: cYE.................cY0    \n"
        "pmulhw 48("RBX"), %%xmm7        # XMM7: cYF.................cY1    \n"
        "movdqa 80("RBX"), %%xmm0        # XMM0: gU constant                \n"
        "pmulhw %%xmm2, %%xmm0           # XMM0: gUL7...............gUL0    \n"
        "movdqa 96("RBX"), %%xmm1        # XMM1: gV constant                \n"
        "pmulhw %%xmm3, %%xmm1           # XMM1: gVL7...............gVL0    \n"
        "paddw %%xmm0, %%xmm1            # XMM1: gE gC gA g8 g6 g4 g2 g0    \n"
        "pmulhw 64("RBX"), %%xmm3        # XMM3: rE rC rA r8 r6 r4 r2 r0    \n"
        "pmulhw 112("RBX"),%%xmm2        # XMM2: bE bC bA b8 b6 b4 b2 b0    \n"
        "movdqa %%xmm3, %%xmm0           # XMM0: rE rC rA r8 r6 r4 r2 r0    \n"
#ifdef SIL_UTILITY_YUV2RGB_SMOOTH_UV
        "movdqa 64("RSI"), %%xmm5        # XMM5: UR7.................UR0    \n"
        "movdqa 80("RSI"), %%xmm3        # XMM3: VR7.................VR0    \n"
        "pmulhw 80("RBX"), %%xmm5        # XMM5: gUR7...............gUR0    \n"
        "movdqa 96("RBX"), %%xmm4        # XMM4: gV constant                \n"
        "pmulhw %%xmm3, %%xmm4           # XMM1: gVL7...............gVL0    \n"
        "paddw  %%xmm5, %%xmm4           # XMM4: gF gD gB g9 g7 g5 g3 g1    \n"
        "movdqa 64("RSI"), %%xmm5        # XMM5: UR7.................UR0    \n"
        "pmulhw 64("RBX"), %%xmm3        # XMM3: rF rD rB r9 r7 r5 r3 r1    \n"
        "pmulhw 112("RBX"),%%xmm5        # XMM5: bF bD bB b9 b7 b5 b3 b1    \n"
#else  // !SIL_UTILITY_YUV2RGB_SMOOTH_UV
        "movdqa %%xmm1, %%xmm4           # XMM4: gE gC gA g8 g6 g4 g2 g0    \n"
        "movdqa %%xmm2, %%xmm5           # XMM5: bE bC bA b8 b6 b4 b2 b0    \n"
#endif
        "# Add intermediate results and round/shift to get R/G/B values     \n"
        "paddw 128("RBX"), %%xmm6 # Add rounding value (0.5 @ 8.4 fixed)    \n"
        "paddw 128("RBX"), %%xmm7                                           \n"
        "paddw %%xmm6, %%xmm0            # XMM0: RE RC RA R8 R6 R4 R2 R0    \n"
        "psraw $4, %%xmm0                # Shift back to 8.0 fixed          \n"
        "paddw %%xmm6, %%xmm1            # XMM1: GE GC GA G8 G6 G4 G2 G0    \n"
        "psraw $4, %%xmm1                                                   \n"
        "paddw %%xmm6, %%xmm2            # XMM2: BE BC BA B8 B6 B4 B2 B0    \n"
        "psraw $4, %%xmm2                                                   \n"
        "paddw %%xmm7, %%xmm3            # XMM3: RF RD RB R9 R7 R5 R3 R1    \n"
        "psraw $4, %%xmm3                                                   \n"
        "paddw %%xmm7, %%xmm4            # XMM4: GF GD GB G9 G7 G5 G3 G1    \n"
        "psraw $4, %%xmm4                                                   \n"
        "paddw %%xmm7, %%xmm5            # XMM5: BF BD BB B9 B7 B5 B3 B1    \n"
        "psraw $4, %%xmm5                                                   \n"
        "# Saturate to 0-255, pack into bytes, interleave, and store        \n"
        "packuswb %%xmm0, %%xmm0         # XMM0: RE.......R0 RE.......R0    \n"
        "packuswb %%xmm1, %%xmm1         # XMM1: GE.......G0 GE.......G0    \n"
        "packuswb %%xmm2, %%xmm2         # XMM2: BE.......B0 BE.......B0    \n"
        "packuswb %%xmm3, %%xmm3         # XMM3: RF.......R1 RF.......R1    \n"
        "packuswb %%xmm4, %%xmm4         # XMM4: GF.......G1 GF.......G1    \n"
        "packuswb %%xmm5, %%xmm5         # XMM5: BF.......B1 BF.......B1    \n"
        "punpcklbw %%xmm3, %%xmm0        # XMM0: RF...................R0    \n"
        "punpcklbw %%xmm4, %%xmm1        # XMM1: GF...................G0    \n"
        "punpcklbw %%xmm5, %%xmm2        # XMM2: BF...................B0    \n"
        "movdqa 144("RBX"),%%xmm3        # XMM3: FF...................FF    \n"
        "movdqa %%xmm0, %%xmm4           # XMM4: RF...................R0    \n"
        "movdqa %%xmm2, %%xmm6           # XMM4: BF...................B0    \n"
        "punpcklbw %%xmm1, %%xmm0        # XMM2: G7 R7.............G0 R0    \n"
        "punpckhbw %%xmm1, %%xmm4        # XMM6: GF RF.............G8 R8    \n"
        "punpcklbw %%xmm3, %%xmm2        # XMM0: FF B7.............FF B0    \n"
        "punpckhbw %%xmm3, %%xmm6        # XMM4: FF BF.............FF B8    \n"
        "movdqa %%xmm0, %%xmm3           # XMM3: G7 R7.............G0 R0    \n"
        "movdqa %%xmm4, %%xmm7           # XMM7: GF RF.............G8 R8    \n"
        "punpcklwd %%xmm2, %%xmm0        # XMM2: FFB3G3R3.......FFB0G0R0    \n"
        "punpckhwd %%xmm2, %%xmm3        # XMM3: FFB7G7R7.......FFB4G4R4    \n"
        "punpcklwd %%xmm6, %%xmm4        # XMM2: FFBBGBRB.......FFB8G8R8    \n"
        "punpckhwd %%xmm6, %%xmm7        # XMM3: FFBFGFRF.......FFBCGCRC    \n"
        "movdqa %%xmm0, ("RDI","RCX",8)                                     \n"
        "movdqa %%xmm3, 16("RDI","RCX",8)                                   \n"
        "movdqa %%xmm4, 32("RDI","RCX",8)                                   \n"
        "movdqa %%xmm7, 48("RDI","RCX",8)                                   \n"
        "# Loop over input row                                              \n"
        "sub $8, "RCX"                                                      \n"
        "jge 0b                                                             \n"
#ifdef SIL_ARCH_X86_32
# if defined(__GNUC__) && !defined(__clang__) && __GNUC__ == 4 && __GNUC_MINOR__ >= 6 && __GNUC_MINOR__ <= 8
        "pop %%edx                                                          \n"
# endif
        "pop %%ebx                                                          \n"
#endif
        : "=c" (dummy)
        : "S" (srcYUV), "m" (*srcYUV), "D" (dest), "c" (width/2-8),
          [yuv_data] "m" (yuv_data)
        : "eax",
#ifndef SIL_ARCH_X86_32
          "ebx",
#endif
#if !defined(SIL_ARCH_X86_32) || !(defined(__GNUC__) && !defined(__clang__) && __GNUC__ == 4 && __GNUC_MINOR__ >= 6 && __GNUC_MINOR__ <= 8)
          "edx",
#endif
          "xmm0", "xmm1", "xmm2", "xmm3", "xmm4", "xmm5", "xmm6", "xmm7"
    );
}

#endif  // SIL_ARCH_X86

/*************************************************************************/
/*************************************************************************/

/*
 * Local variables:
 *   c-file-style: "stroustrup"
 *   c-file-offsets: ((case-label . *) (statement-case-intro . *))
 *   indent-tabs-mode: nil
 * End:
 *
 * vim: expandtab shiftwidth=4:
 */
